from typing import List

from fastapi import HTTPException, status

from id_broker_common.user import user_getter_by_broker_id
from ucsschool.apis.models import OAuth2Token
from ucsschool.apis.utils import auth_manager, get_logger
from ucsschool.kelvin.client import (
    InvalidRequest,
    SchoolClass,
    SchoolClassResource,
    Session,
    User,
    UserResource,
)

from .models import ClassMembers, Student, Teacher

# Extracted to variable to allow dep overrides in FastAPI, which cannot resolve anonymous dependencies
# TODO: import PLUGIN_NAME without circular import error:
token_dep = auth_manager("self_disclosure", OAuth2Token)


def roles_in_school(user: User, school_name: str) -> List[str]:
    """Helper to get only the roles that are defined for an user within a specific school"""
    roles = set()
    for role_string in user.ucsschool_roles:
        role, context_type, context_name = role_string.split(":")
        if context_type == "school" and school_name == context_name:
            roles.add(role)
    return list(roles)


async def extract_members(schoolclass: SchoolClass, session: Session) -> ClassMembers:
    """Helper to get users of a school_class split into teachers and students"""
    logger = get_logger()
    teachers = []
    students = []
    user_name: str
    for user_name in schoolclass.users:
        user = await get_single_user_by_name(user_name, session)
        if not (hasattr(user, "udm_properties") and "brokerID" in user.udm_properties):
            msg = "udm_properties -> brokerID field required (type=value_error.missing)"
            logger.warning(f"user '{user_name}' did not validate: {msg}")
            continue
        roles = roles_in_school(user, schoolclass.school)
        if "student" in roles:
            student = Student(
                user_id=user.udm_properties["brokerID"],
                username=user.name,
                firstname=user.firstname,
                lastname=user.lastname,
            )
            students.append(student)
        if "teacher" in roles:
            teacher = Teacher(user_id=user.udm_properties["brokerID"], username=user.name)
            teachers.append(teacher)
    return ClassMembers(students=students, teachers=teachers)


async def get_single_user_by_broker_id(broker_id: str, session: Session) -> User:
    """Search for a single user by brokerID. Raise HTTPExceptions if not exactly a single user is found."""
    users = await user_getter_by_broker_id(
        broker_id, session
    )  # This will also raise an HTTPException if '*' is in broker_id
    if len(users) > 1:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Did find more than one user with broker-id {broker_id!r}.",
        )
    elif not users:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No user was found with broker-id {broker_id!r}.",
        )
    return users[0]


async def get_single_user_by_name(name: str, session: Session) -> User:
    """Search for a single user by name."""
    return await UserResource(session=session).get(name=name)


async def schoolclass_getter(school: str, name: str, session: Session) -> SchoolClass:
    """Search for schoolclass. Raise HTTPExceptions if not exactly a single result is returned."""
    try:
        res = [
            obj async for obj in SchoolClassResource(session=session).search(school=school, name=name)
        ]
    except InvalidRequest:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No class was found for {school}-{name}.",
        )

    if len(res) > 1:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=f"Found more then one class for {school}-{name}.",
        )
    elif not res:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No class was found for {school}-{name}.",
        )
    return res[0]
