from unittest.mock import AsyncMock

import pytest
from fastapi.testclient import TestClient

from id_broker_common.kelvin import kelvin_session
from self_disclosure_plugin import plugin
from self_disclosure_plugin.dependencies import extract_members, roles_in_school, token_dep
from ucsschool.apis.models import OAuth2Token
from ucsschool.apis.opa import opa_instance
from ucsschool.kelvin.client import SchoolClass

from ..utils import DictObject, get_preconfigured_session_mock


@pytest.fixture(scope="session")
def client(app, dependency_overrides, fake_opa, async_dict):
    # FastAPI Depends() overrides for complete session
    app.include_router(plugin.router)
    overrides = {
        opa_instance: fake_opa,
        token_dep: lambda: OAuth2Token(iss="", sub="", aud=[], exp=0, client_id="", iat=0, jti=""),
        kelvin_session: get_preconfigured_session_mock(),
    }
    dependency_overrides(overrides)
    return TestClient(app)


kelvin_testdata = DictObject()

kelvin_testdata.student1 = DictObject(
    __name__="student1",  # needed for parametrize
    udm_properties=dict(brokerID="stu1@test"),
    name="student1",
    roles=["student"],
    firstname="stu1",
    lastname="dent",
    school_classes=dict(demoschool=["class1"]),
    school="demoschool",
    ucsschool_roles=["student:school:demoschool"],
)

kelvin_testdata.student2 = DictObject(
    __name__="student2",  # needed for parametrize
    udm_properties=dict(brokerID="stu2@test"),
    name="student2",
    roles=["student"],
    firstname="stu2",
    lastname="dent",
    school_classes=dict(demoschool=["class1"]),
    school="demoschool",
    ucsschool_roles=["student:school:demoschool"],
)

kelvin_testdata.student3 = DictObject(
    __name__="student3",  # needed for parametrize
    udm_properties=dict(brokerID="stu3@test"),
    name="student3",
    roles=["student"],
    firstname="stu3",
    lastname="dent",
    school_classes=dict(demoschool=["class2"]),
    school="demoschool",
    ucsschool_roles=["student:school:demoschool"],
)

kelvin_testdata.student4 = DictObject(
    __name__="student4",  # needed for parametrize
    udm_properties=dict(),  # no brokerid!
    name="student4",
    roles=["student"],
    firstname="stu4",
    lastname="dent",
    school_classes=dict(demoschool=["class2"]),
    school="demoschool",
    ucsschool_roles=["student:school:demoschool"],
)
kelvin_testdata.teacher1 = DictObject(
    __name__="teacher1",  # needed for parametrize
    udm_properties=dict(brokerID="tea1@test"),
    name="teacher1",
    roles=["teacher"],
    firstname="tea1",
    lastname="cher",
    school_classes=dict(demoschool=["class1"]),
    school="demoschool",
    ucsschool_roles=["teacher:school:demoschool"],
)
kelvin_testdata.class1 = DictObject(
    name="class1",
    users=[
        kelvin_testdata.student1.name,
        kelvin_testdata.student2.name,
        kelvin_testdata.teacher1.name,
    ],
    school="demoschool",
)
kelvin_testdata.class2 = DictObject(
    name="class2",
    users=[kelvin_testdata.student3.name, kelvin_testdata.student4.name],
    school="demoschool",
)


def test_user_metadata_correct(client, monkeypatch):
    users = [kelvin_testdata.student1]
    monkeypatch.setattr(plugin, "user_getter_by_broker_id", AsyncMock(return_value=users))
    response = client.get(f"/users/{kelvin_testdata.student1.udm_properties['brokerID']}/metadata")
    assert response.status_code == 200
    expected_user = dict(
        user_id=kelvin_testdata.student1.udm_properties["brokerID"],
        username=kelvin_testdata.student1.name,
        firstname=kelvin_testdata.student1.firstname,
        lastname=kelvin_testdata.student1.lastname,
        type=kelvin_testdata.student1.roles[0],
        school_id=kelvin_testdata.student1.school,
    )
    assert response.json() == expected_user


def test_user_groups_correct(client, monkeypatch):
    class_members = [
        kelvin_testdata.teacher1,
        kelvin_testdata.student1,
        kelvin_testdata.student2,
    ]
    monkeypatch.setattr(
        plugin,
        "get_single_user_by_broker_id",
        AsyncMock(return_value=kelvin_testdata.student1),
    )
    monkeypatch.setattr(
        plugin,
        "get_single_user_by_name",
        AsyncMock(side_effect=class_members),
    )
    monkeypatch.setattr(
        plugin,
        "schoolclass_getter",
        AsyncMock(return_value=kelvin_testdata.class1),
    )
    response = client.get(f"/users/{kelvin_testdata.teacher1.udm_properties['brokerID']}/groups")
    assert response.status_code == 200
    expected_groups = dict(
        groups=[
            {
                "group_id": f"{kelvin_testdata.class1.school}-{kelvin_testdata.class1.name}",
                "name": f"{kelvin_testdata.class1.school}-{kelvin_testdata.class1.name}",
                "student_count": 2,
            }
        ]
    )
    assert response.json() == expected_groups


@pytest.mark.skip("Test does not work and need to be fixed when proper pseudonyms are implemented")
def test_group_users_correct(client, monkeypatch):
    users = [
        kelvin_testdata.teacher1,
        kelvin_testdata.student1,
        kelvin_testdata.student2,
    ]
    school_classes = kelvin_testdata.class1
    monkeypatch.setattr(plugin, "schoolclass_getter", AsyncMock(return_value=school_classes))
    monkeypatch.setattr(plugin, "get_single_user_by_broker_id", AsyncMock(side_effect=users))
    response = client.get(f"/groups/{kelvin_testdata.class1.school}-{kelvin_testdata.class1.name}/users")
    assert response.status_code == 200
    expected_group = dict(
        students=[
            dict(
                user_id=kelvin_testdata.student1.udm_properties["brokerID"],
                username=kelvin_testdata.student1.name,
                firstname=kelvin_testdata.student1.firstname,
                lastname=kelvin_testdata.student1.lastname,
            ),
            dict(
                user_id=kelvin_testdata.student2.udm_properties["brokerID"],
                username=kelvin_testdata.student2.name,
                firstname=kelvin_testdata.student2.firstname,
                lastname=kelvin_testdata.student2.lastname,
            ),
        ],
        teachers=[
            dict(
                user_id=kelvin_testdata.teacher1.udm_properties["brokerID"],
                username=kelvin_testdata.teacher1.name,
            )
        ],
    )
    assert response.json() == expected_group


@pytest.mark.asyncio
async def test_extract_members(monkeypatch):
    users = [
        kelvin_testdata.teacher1,
        kelvin_testdata.student1,
        kelvin_testdata.student2,
    ]
    monkeypatch.setattr(plugin, "get_single_user_by_name", AsyncMock(side_effect=users))
    classmembers = await extract_members(
        SchoolClass(**kelvin_testdata.class1), get_preconfigured_session_mock()
    )
    assert classmembers.dict() == dict(
        students=[
            dict(
                user_id=kelvin_testdata.student1.udm_properties["brokerID"],
                username=kelvin_testdata.student1.name,
                firstname=kelvin_testdata.student1.firstname,
                lastname=kelvin_testdata.student1.lastname,
            ),
            dict(
                user_id=kelvin_testdata.student2.udm_properties["brokerID"],
                username=kelvin_testdata.student2.name,
                firstname=kelvin_testdata.student2.firstname,
                lastname=kelvin_testdata.student2.lastname,
            ),
        ],
        teachers=[
            dict(
                user_id=kelvin_testdata.teacher1.udm_properties["brokerID"],
                username=kelvin_testdata.teacher1.name,
            )
        ],
    )


@pytest.mark.parametrize("user", [kelvin_testdata.student1, kelvin_testdata.teacher1])
def test_roles_in_school(user):
    roles = roles_in_school(user, user.school)
    assert set(roles) == set(user.roles)


@pytest.mark.asyncio
async def test_extract_members_filtered(monkeypatch, caplog):
    """Test if extract filters users without brokerID and logs a warning for them"""
    users = [kelvin_testdata.student3, kelvin_testdata.student4]
    monkeypatch.setattr(plugin, "get_single_user_by_name", AsyncMock(side_effect=users))
    classmembers = await extract_members(
        SchoolClass(**kelvin_testdata.class2), get_preconfigured_session_mock()
    )
    assert classmembers.dict() == dict(
        students=[
            dict(
                user_id=kelvin_testdata.student3.udm_properties["brokerID"],
                username=kelvin_testdata.student3.name,
                firstname=kelvin_testdata.student3.firstname,
                lastname=kelvin_testdata.student3.lastname,
            ),
        ],
        teachers=[],
    )
    assert "udm_properties -> brokerID field required (type=value_error.missing)" in caplog.text


def test_user_groups_filtered(client, monkeypatch, caplog):
    class_members = [
        kelvin_testdata.student3,
        kelvin_testdata.student4,
    ]
    monkeypatch.setattr(
        plugin,
        "get_single_user_by_broker_id",
        AsyncMock(return_value=kelvin_testdata.student3),
    )
    monkeypatch.setattr(
        plugin,
        "get_single_user_by_name",
        AsyncMock(side_effect=class_members),
    )
    monkeypatch.setattr(
        plugin,
        "schoolclass_getter",
        AsyncMock(return_value=kelvin_testdata.class2),
    )
    response = client.get(f"/users/{kelvin_testdata.student3.udm_properties['brokerID']}/groups")
    assert response.status_code == 200
    expected_groups = dict(
        groups=[
            {
                "group_id": f"{kelvin_testdata.class2.school}-{kelvin_testdata.class2.name}",
                "name": f"{kelvin_testdata.class2.school}-{kelvin_testdata.class2.name}",
                "student_count": 1,
            }
        ]
    )
    assert response.json() == expected_groups
    assert "udm_properties -> brokerID field required (type=value_error.missing)" in caplog.text


@pytest.mark.skip("Test does not work and need to be fixed when proper pseudonyms are implemented")
def test_group_users_filtered(client, monkeypatch, caplog):
    users = [kelvin_testdata.student3, kelvin_testdata.student4]
    monkeypatch.setattr(
        plugin,
        "schoolclass_getter",
        AsyncMock(return_value=kelvin_testdata.class2),
    )
    monkeypatch.setattr(plugin, "get_single_user_by_broker_id", AsyncMock(side_effect=users))
    response = client.get(f"/groups/{kelvin_testdata.class2.school}-{kelvin_testdata.class2.name}/users")
    assert response.status_code == 200
    expected_group = dict(
        students=[
            dict(
                user_id=kelvin_testdata.student3.udm_properties["brokerID"],
                username=kelvin_testdata.student3.name,
                firstname=kelvin_testdata.student3.firstname,
                lastname=kelvin_testdata.student3.lastname,
            ),
        ],
        teachers=[],
    )
    assert response.json() == expected_group
    assert "udm_properties -> brokerID field required (type=value_error.missing)" in caplog.text
